#!/usr/bin/env python

"""
Phrase iterators providing navigation over common positions for a number of
different terms.

Copyright (C) 2009, 2010 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from itermerge import itermerge
from bisect import insort_right

class CommonIterator(itermerge):

    """
    Iteration over many terms, driving the search for term co-occurrences using
    the least frequent term.
    """

    def _add_seq(self, sequence, i):

        "Store the details of the given 'sequence' at position 'i'."

        insort_right(self.iters, (len(sequence), i, iter(sequence)))

    def next(self):

        """
        Return a tuple containing a document identifier and a list of position
        lists for all terms.
        """

        if self.iters:
            while 1:

                # Get the next document for the lowest frequency term.

                freq, i, it = self.iters[0]
                doc, positions = it.next()
                values = [(i, positions)]

                # Attempt to find the other terms in this document.

                for freq, i, it in self.iters[1:]:
                    positions = it.from_document(doc)

                    # Insert position details if appropriate.

                    if positions is not None:
                        values.append((i, positions))

                    # Otherwise, reject this document.

                    else:
                        break
                else:
                    values.sort()
                    return doc, [positions for (i, positions) in values]
        else:
            raise StopIteration

    def close(self):
        for freq, i, it in self.iters:
            if hasattr(it, "close"):
                it.close()
        self.iters = []

class PhraseIterator(CommonIterator):

    "Phrase iteration using the phrase filter."

    def __init__(self, sequences, filter=None):
        CommonIterator.__init__(self, sequences)
        self.current_doc = None
        self.current_positions = None
        self.filter = filter or PhraseFilter

    def next(self):

        """
        Return a tuple containing a document identifier and a list of term
        positions.
        """

        while 1:
            if self.current_doc is None:
                self.current_doc, all_positions = CommonIterator.next(self)

                # Handle incomplete phrases.

                try:
                    self.current_positions = self.filter(all_positions)
                except StopIteration:
                    self.current_doc = None
                    continue

            # Return new phrases.

            try:
                return self.current_doc, self.current_positions.next()
            except StopIteration:
                self.current_doc = None

class PhraseFilter(itermerge):

    "Filter phrase suggestions according to position information."

    def _add_seq(self, sequence, i):

        "Store the details of the given 'sequence' at position 'i'."

        it = iter(sequence)
        self._add_next(it.next, i)

    def _add_next(self, next, i):

        """
        Store the current value for an iterator, alongside the means of
        getting the next value - the 'next' method - together with the
        iterator's position 'i'.
        """

        # Allow StopIteration to be raised.

        insort_right(self.iters, (next(), i, next))

    def next(self):
        if self.iters:
            while 1:
                current, first_token, next = self.iters[0]
                values = [current]
                last = current
                last_token = first_token

                # Find a sequence of positions providing a phrase.

                for current, current_token, _next in self.iters[1:]:
                    if not self.is_phrase_position(last, last_token, current, current_token):
                        break
                    values.append(current)
                    last = current
                    last_token = current_token
                else:
                    del self.iters[0]

                    # Handle future end of iteration.

                    try:
                        self._add_next(next, first_token)
                    except StopIteration:
                        self.iters = []

                    return values

                del self.iters[0]
                self._add_next(next, first_token)
        else:
            raise StopIteration

    def is_phrase_position(self, last, last_token, current, current_token):
        if current_token <= last_token:
            return 0

        # NOTE: For position sequences, assume that the first value is the token
        # NOTE: index/position.

        if isinstance(last, (list, tuple)):
            return current[0] - last[0] == 1
        else:
            return current - last == 1

# vim: tabstop=4 expandtab shiftwidth=4
